@@ -1,3 +1,5 @@
+
+
 /*
  * This is free software; you can redistribute it and/or modify it under
  * the terms of the GNU Library General Public License as published by
@@ -23,7 +25,6 @@
 #include <tilda_window.h>
 #include <tilda_terminal.h>
 #include <key_grabber.h>
-#include <translation.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -31,26 +32,28 @@
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
 #include <glib-object.h>
+#include <glib.h>
+#include <glib/gi18n.h>
 #include <vte/vte.h>
 
 static void
 tilda_window_setup_alpha_mode (tilda_window *tw)
 {
     GdkScreen *screen;
-    GdkColormap *colormap;
+    GdkVisual *visual;
 
     screen = gtk_widget_get_screen (GTK_WIDGET (tw->window));
-    colormap = gdk_screen_get_rgba_colormap (screen);
-    if (colormap != NULL && gdk_screen_is_composited (screen))
-    {
+    visual = gdk_screen_get_rgba_visual (screen);
+    if (visual == NULL) {
+        visual = gdk_screen_get_system_visual (screen);
+    }
+    if (visual != NULL && gdk_screen_is_composited (screen)) {
         /* Set RGBA colormap if possible so VTE can use real alpha
          * channels for transparency. */
 
-        gtk_widget_set_colormap(GTK_WIDGET (tw->window), colormap);
+        gtk_widget_set_visual (GTK_WIDGET (tw->window), visual);
         tw->have_argb_visual = TRUE;
-    }
-    else
-    {
+    } else {
         tw->have_argb_visual = FALSE;
     }
 }
@@ -87,7 +90,7 @@ void tilda_window_close_current_tab (tilda_window *tw)
     DEBUG_ASSERT (tw != NULL);
 
     gint pos = gtk_notebook_get_current_page (GTK_NOTEBOOK (tw->notebook));
-    tilda_window_close_tab (tw, pos);
+    tilda_window_close_tab (tw, pos, FALSE);
 }
 
 
@@ -114,6 +117,23 @@ gint tilda_window_set_tab_position (tilda_window *tw, enum notebook_tab_position
     return 0;
 }
 
+
+gint toggle_fullscreen_cb (tilda_window *tw)
+{
+    DEBUG_FUNCTION ("toggle_fullscreen_cb");
+    DEBUG_ASSERT (tw != NULL);
+    
+	if (tw->fullscreen != TRUE) {
+		tw->fullscreen = TRUE;
+		gtk_window_fullscreen (GTK_WINDOW (tw->window));
+	}
+	else {
+		gtk_window_unfullscreen (GTK_WINDOW (tw->window));
+		tw->fullscreen = FALSE;
+	}
+}
+
+
 static gint next_tab (tilda_window *tw)
 {
     DEBUG_FUNCTION ("next_tab");
@@ -122,13 +142,14 @@ static gint next_tab (tilda_window *tw)
     int num_pages;
     int current_page;
 
+    /* If we are on the last page, go to first page */
     num_pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (tw->notebook));
     current_page = gtk_notebook_get_current_page (GTK_NOTEBOOK (tw->notebook));
 
-    if (num_pages != (current_page + num_pages))
-      gtk_notebook_next_page (GTK_NOTEBOOK (tw->notebook));
+    if ((num_pages - 1) == current_page)
+      gtk_notebook_set_current_page (GTK_NOTEBOOK (tw->notebook), 0);
     else
-      gtk_notebook_set_current_page (GTK_NOTEBOOK (tw->notebook), num_pages-1);
+      gtk_notebook_next_page (GTK_NOTEBOOK (tw->notebook));
 
     // It worked. Having this return true makes the callback not carry the
     // keystroke into the vte terminal widget.
@@ -146,10 +167,69 @@ static gint prev_tab (tilda_window *tw)
     num_pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (tw->notebook));
     current_page = gtk_notebook_get_current_page (GTK_NOTEBOOK (tw->notebook));
 
-    if ((num_pages-1) != current_page)
-      gtk_notebook_prev_page (GTK_NOTEBOOK (tw->notebook));
+    if (current_page == 0)
+        gtk_notebook_set_current_page (GTK_NOTEBOOK (tw->notebook), (num_pages -1));
     else
-      gtk_notebook_set_current_page (GTK_NOTEBOOK (tw->notebook), 0);
+      gtk_notebook_prev_page (GTK_NOTEBOOK (tw->notebook));
+
+    // It worked. Having this return true makes the callback not carry the
+    // keystroke into the vte terminal widget.
+    return TRUE;
+}
+
+static gint move_tab_left (tilda_window *tw)
+{
+    DEBUG_FUNCTION ("move_tab_left");
+    DEBUG_ASSERT (tw != NULL);
+
+    int num_pages;
+    int current_page_index;
+    int new_page_index;
+    GtkWidget* current_page;
+
+    num_pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (tw->notebook));
+    current_page_index = gtk_notebook_get_current_page (GTK_NOTEBOOK (tw->notebook));
+    current_page = gtk_notebook_get_nth_page (GTK_NOTEBOOK (tw->notebook),
+                                              current_page_index);
+
+    if (num_pages > 1) {
+      if (current_page_index < num_pages - 1) {
+        // Move current page one to the left
+        new_page_index = current_page_index + 1;
+      } else {
+        // Current page is at beginning: move to end
+        new_page_index = 0;
+      }
+
+      gtk_notebook_reorder_child (GTK_NOTEBOOK (tw->notebook), current_page,
+                                  new_page_index);
+    }
+
+    // It worked. Having this return true makes the callback not carry the
+    // keystroke into the vte terminal widget.
+    return TRUE;
+}
+
+static gint move_tab_right (tilda_window *tw)
+{
+    DEBUG_FUNCTION ("move_tab_right");
+    DEBUG_ASSERT (tw != NULL);
+
+    int num_pages;
+    int current_page_index;
+    GtkWidget* current_page;
+
+    num_pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (tw->notebook));
+    current_page_index = gtk_notebook_get_current_page (GTK_NOTEBOOK (tw->notebook));
+    current_page = gtk_notebook_get_nth_page (GTK_NOTEBOOK (tw->notebook),
+                                              current_page_index);
+
+    if (num_pages > 1) {
+      // Move current page one to the right. This automatically wraps
+      // when current_page_index - 1 is negative.
+      gtk_notebook_reorder_child (GTK_NOTEBOOK (tw->notebook), current_page,
+                                  current_page_index - 1);
+    }
 
     // It worked. Having this return true makes the callback not carry the
     // keystroke into the vte terminal widget.
@@ -164,20 +244,109 @@ static void focus_term (GtkWidget *widget, gpointer data)
 
     GList *list;
     GtkWidget *box;
-    GtkWidget *n = GTK_WIDGET (data);
+    tilda_window *tw = TILDA_WINDOW(data);
+    GtkWidget *n = GTK_WIDGET (tw->notebook);
 
     box = gtk_notebook_get_nth_page (GTK_NOTEBOOK(n), gtk_notebook_get_current_page(GTK_NOTEBOOK(n)));
-    list = gtk_container_get_children (GTK_CONTAINER(box));
-    gtk_widget_grab_focus (list->data);
+    if (box != NULL) {
+        list = gtk_container_get_children (GTK_CONTAINER(box));
+        gtk_widget_grab_focus (list->data);
+    }
 }
 
-static void focus_out_event_cb (GtkWidget *widget, gpointer data)
+static gboolean auto_hide_tick(gpointer data)
+{
+    DEBUG_FUNCTION ("auto_hide_tick");
+    DEBUG_ASSERT (data != NULL);
+    
+    tilda_window *tw = TILDA_WINDOW(data);
+    tw->auto_hide_current_time += tw->timer_resolution;
+    if ((tw->auto_hide_current_time >= tw->auto_hide_max_time) || tw->current_state == UP)
+    {
+        pull(tw, PULL_UP);
+        tw->auto_hide_tick_handler = 0;
+        return FALSE;
+    }
+    return TRUE;
+}
+
+/* Start auto hide tick */
+static void start_auto_hide_tick(tilda_window *tw)
+{
+    // If the delay is less or equal to 1000ms then the timer is not precise
+    // enough, because it takes already about 200ms to register it, so we
+    // rather sleep for the given amount of time.
+    const int MAX_SLEEP_TIME = 1000;
+
+    if ((tw->auto_hide_tick_handler == 0) && (tw->disable_auto_hide == FALSE)) {
+        /* If there is not timer registered yet, then the auto_hide_tick_handler
+         * has a value of zero. Next we need to make sure that the auto hide
+         * max time is greater then zero, or else we can pull up immediately,
+         * without the trouble of registering a timer.
+         */
+        if (tw->auto_hide_max_time > MAX_SLEEP_TIME) {
+            tw->auto_hide_current_time = 0;
+            tw->auto_hide_tick_handler = g_timeout_add(tw->timer_resolution, auto_hide_tick, tw);
+        } else if (tw->auto_hide_max_time <= MAX_SLEEP_TIME) {
+            // auto_hide_max_time is in milli seconds, so we need to convert to
+            // microseconds by multiplying with 1000
+            g_usleep (tw->auto_hide_max_time * 1000);
+            pull(tw, PULL_UP);
+        } else {
+            pull(tw, PULL_UP);
+        }
+    }
+}
+
+/* Stop auto hide tick */
+static void stop_auto_hide_tick(tilda_window *tw)
+{
+    if (tw->auto_hide_tick_handler != 0)
+    {
+        g_source_remove(tw->auto_hide_tick_handler);
+        tw->auto_hide_tick_handler = 0;
+    }
+}
+
+static void mouse_enter (GtkWidget *widget, GdkEvent *event, gpointer data)
+{
+    DEBUG_FUNCTION ("mouse_enter");
+    DEBUG_ASSERT (data != NULL);
+    DEBUG_ASSERT (widget != NULL);
+    
+    tilda_window *tw = TILDA_WINDOW(data);
+    stop_auto_hide_tick(tw);
+    if (tw->disable_auto_hide == FALSE)
+        tilda_window_set_active(tw);
+}
+ 
+static void mouse_leave (GtkWidget *widget, GdkEvent *event, gpointer data)
+{
+    DEBUG_FUNCTION ("mouse_leave");
+    DEBUG_ASSERT (data != NULL);
+    DEBUG_ASSERT (widget != NULL);
+    
+    GdkEventCrossing *ev = (GdkEventCrossing*)event;
+    tilda_window *tw = TILDA_WINDOW(data);
+    
+    if ((ev->mode != GDK_CROSSING_NORMAL) || (tw->auto_hide_on_mouse_leave == FALSE))
+        return;
+    
+    start_auto_hide_tick(tw);
+}
+
+static void focus_out_event_cb (GtkWidget *widget, GdkEvent *event, gpointer data)
 {
     DEBUG_FUNCTION ("focus_out_event_cb");
     DEBUG_ASSERT (data != NULL);
     DEBUG_ASSERT (widget != NULL);
-
-    //printf ("hide()\n");
+    
+    tilda_window *tw = TILDA_WINDOW(data);
+    
+    if (tw->auto_hide_on_focus_lost == FALSE)
+        return;
+    
+    start_auto_hide_tick(tw);
 }
 
 static void goto_tab (tilda_window *tw, guint i)
@@ -195,7 +364,7 @@ static gboolean goto_tab_generic (tilda_window *tw, gint tab_number)
 
     if (g_list_length (tw->terms) > (tab_number-1))
     {
-        goto_tab (tw, g_list_length (tw->terms) - tab_number);
+        goto_tab (tw, tab_number - 1);
         return TRUE;
     }
 
@@ -252,73 +421,62 @@ static gint cpaste (tilda_window *tw)
     return TRUE;
 }
 
-static gint tilda_window_setup_keyboard_accelerators (tilda_window *tw)
+/* Tie a single keyboard shortcut to a callback function */
+gint tilda_add_config_accelerator(const gchar* key, GCallback callback_func, tilda_window *tw)
 {
-    GtkAccelGroup *accel_group;
+    guint accel_key;
+    GdkModifierType accel_mods;
     GClosure *temp;
-
+ 
+    gtk_accelerator_parse (config_getstr(key), &accel_key, &accel_mods);
+    if (! ((accel_key == 0) && (accel_mods == 0)) )  // make sure it parsed properly
+    {
+        temp = g_cclosure_new_swap (callback_func, tw, NULL);
+        gtk_accel_group_connect (tw->accel_group, accel_key, accel_mods , GTK_ACCEL_VISIBLE, temp);
+    }
+ 
+    return 0;
+}
+ 
+gint tilda_window_setup_keyboard_accelerators (tilda_window *tw)
+{
+ 
+    /* If we already have an tw->accel_group (which would happen if we're redefining accelerators in the config window)
+       we want to remove it before creating a new one. */
+    if (tw->accel_group != NULL)
+        gtk_window_remove_accel_group (GTK_WINDOW (tw->window), tw->accel_group);
+ 
     /* Create Accel Group to add key codes for quit, next, prev and new tabs */
-    accel_group = gtk_accel_group_new ();
-    gtk_window_add_accel_group (GTK_WINDOW (tw->window), accel_group);
-
-    /* Exit on <Ctrl><Shift>q */
-    temp = g_cclosure_new_swap (G_CALLBACK(gtk_main_quit), tw, NULL);
-    gtk_accel_group_connect (accel_group, 'q', GDK_CONTROL_MASK | GDK_SHIFT_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    /* Go to Next Tab on <Ctrl>Page_Down */
-    temp = g_cclosure_new_swap (G_CALLBACK(next_tab), tw, NULL);
-    gtk_accel_group_connect (accel_group, GDK_Page_Down, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    /* Go to Prev Tab on <Ctrl>Page_Up */
-    temp = g_cclosure_new_swap (G_CALLBACK(prev_tab), tw, NULL);
-    gtk_accel_group_connect (accel_group, GDK_Page_Up, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    /* Add New Tab on <Ctrl><Shift>t */
-    temp = g_cclosure_new_swap (G_CALLBACK(tilda_window_add_tab), tw, NULL);
-    gtk_accel_group_connect (accel_group, 't', GDK_CONTROL_MASK | GDK_SHIFT_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    /* Close Current Tab on <Ctrl><Shift>w */
-    temp = g_cclosure_new_swap (G_CALLBACK(tilda_window_close_current_tab), tw, NULL);
-    gtk_accel_group_connect (accel_group, 'w', GDK_CONTROL_MASK | GDK_SHIFT_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    /* Goto Tab # */
+    tw->accel_group = gtk_accel_group_new ();
+    gtk_window_add_accel_group (GTK_WINDOW (tw->window), tw->accel_group);
+ 
+    /* Set up keyboard shortcuts for Exit, Next Tab, Previous Tab,
+       Move Tab, Add Tab, Close Tab, Copy, and Paste using key
+       combinations defined in the config. */
+    tilda_add_config_accelerator("quit_key",         G_CALLBACK(gtk_main_quit),                  tw);
+    tilda_add_config_accelerator("nexttab_key",      G_CALLBACK(next_tab),                       tw);
+    tilda_add_config_accelerator("prevtab_key",      G_CALLBACK(prev_tab),                       tw);
+    tilda_add_config_accelerator("movetableft_key",  G_CALLBACK(move_tab_left),                  tw);
+    tilda_add_config_accelerator("movetabright_key", G_CALLBACK(move_tab_right),                 tw);
+    tilda_add_config_accelerator("addtab_key",       G_CALLBACK(tilda_window_add_tab),           tw);
+    tilda_add_config_accelerator("closetab_key",     G_CALLBACK(tilda_window_close_current_tab), tw);
+    tilda_add_config_accelerator("copy_key",         G_CALLBACK(ccopy),                          tw);
+    tilda_add_config_accelerator("paste_key",        G_CALLBACK(cpaste),                         tw);
+    tilda_add_config_accelerator("fullscreen_key",   G_CALLBACK(toggle_fullscreen_cb),           tw);
+ 
+    /* Set up keyboard shortcuts for Goto Tab # using key combinations defined in the config*/
     /* Know a better way? Then you do. */
-    temp = g_cclosure_new_swap (G_CALLBACK(goto_tab_1), tw, NULL);
-    gtk_accel_group_connect (accel_group, '1', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(goto_tab_2), tw, NULL);
-    gtk_accel_group_connect (accel_group, '2', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(goto_tab_3), tw, NULL);
-    gtk_accel_group_connect (accel_group, '3', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(goto_tab_4), tw, NULL);
-    gtk_accel_group_connect (accel_group, '4', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(goto_tab_5), tw, NULL);
-    gtk_accel_group_connect (accel_group, '5', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(goto_tab_6), tw, NULL);
-    gtk_accel_group_connect (accel_group, '6', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(goto_tab_7), tw, NULL);
-    gtk_accel_group_connect (accel_group, '7', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(goto_tab_8), tw, NULL);
-    gtk_accel_group_connect (accel_group, '8', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(goto_tab_9), tw, NULL);
-    gtk_accel_group_connect (accel_group, '9', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(goto_tab_10), tw, NULL);
-    gtk_accel_group_connect (accel_group, '0', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(ccopy), tw, NULL);
-    gtk_accel_group_connect (accel_group, 'c', GDK_CONTROL_MASK | GDK_SHIFT_MASK, GTK_ACCEL_VISIBLE, temp);
-
-    temp = g_cclosure_new_swap (G_CALLBACK(cpaste), tw, NULL);
-    gtk_accel_group_connect (accel_group, 'v', GDK_CONTROL_MASK | GDK_SHIFT_MASK, GTK_ACCEL_VISIBLE, temp);
-
+    tilda_add_config_accelerator("gototab_1_key",  G_CALLBACK(goto_tab_1),  tw);
+    tilda_add_config_accelerator("gototab_2_key",  G_CALLBACK(goto_tab_2),  tw);
+    tilda_add_config_accelerator("gototab_3_key",  G_CALLBACK(goto_tab_3),  tw);
+    tilda_add_config_accelerator("gototab_4_key",  G_CALLBACK(goto_tab_4),  tw);
+    tilda_add_config_accelerator("gototab_5_key",  G_CALLBACK(goto_tab_5),  tw);
+    tilda_add_config_accelerator("gototab_6_key",  G_CALLBACK(goto_tab_6),  tw);
+    tilda_add_config_accelerator("gototab_7_key",  G_CALLBACK(goto_tab_7),  tw);
+    tilda_add_config_accelerator("gototab_8_key",  G_CALLBACK(goto_tab_8),  tw);
+    tilda_add_config_accelerator("gototab_9_key",  G_CALLBACK(goto_tab_9),  tw);
+    tilda_add_config_accelerator("gototab_10_key", G_CALLBACK(goto_tab_10), tw);
+    
     return 0;
 }
 
@@ -340,6 +498,13 @@ static gint tilda_window_set_icon (tilda_window *tw, gchar *filename)
     return 0;
 }
 
+static gboolean delete_event_callback (GtkWidget *widget,
+                                GdkEvent  *event,
+                                gpointer   user_data)
+{
+    gtk_main_quit ();
+}
+
 /**
  * tilda_window_init ()
  *
@@ -359,6 +524,8 @@ tilda_window *tilda_window_init (const gchar *config_file, const gint instance)
     DEBUG_ASSERT (instance >= 0);
 
     tilda_window *tw;
+    GtkCssProvider *provider;
+    GtkStyleContext *style_context;
 
     tw = g_malloc (sizeof(tilda_window));
 
@@ -368,14 +535,23 @@ tilda_window *tilda_window_init (const gchar *config_file, const gint instance)
     /* Set the instance number */
     tw->instance = instance;
 
-    /* Get the user's home directory */
-    tw->home_dir = g_strdup (g_get_home_dir ());
-
     /* Set the config file */
     tw->config_file = g_strdup (config_file);
 
     /* Create the main window */
     tw->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+    
+    /* Generic timer resolution */
+    tw->timer_resolution = config_getint("timer_resolution");
+    
+    /* Auto hide support */
+    tw->auto_hide_tick_handler = 0;
+    tw->auto_hide_max_time = config_getint("auto_hide_time");
+    tw->auto_hide_on_mouse_leave = config_getbool("auto_hide_on_mouse_leave");
+    tw->auto_hide_on_focus_lost = config_getbool("auto_hide_on_focus_lost");
+    tw->disable_auto_hide = FALSE;
+
+	tw->fullscreen = FALSE;
 
     /* Set up all window properties */
     if (config_getbool ("pinned"))
@@ -389,16 +565,48 @@ tilda_window *tilda_window_init (const gchar *config_file, const gint instance)
     tilda_window_setup_alpha_mode (tw);
 
     /* Add keyboard accelerators */
+    tw->accel_group = NULL; /* We can redefine the accelerator group from the wizard; this shows that it's our first time defining it. */
     tilda_window_setup_keyboard_accelerators (tw);
 
     /* Create the notebook */
     tw->notebook = gtk_notebook_new ();
 
-    /* Set up all notebook properties */
+    /* Here we setup the CSS settings for the GtkNotebook.
+     * If the option "Show notebook border" in the preferences is not
+     * checked. Then we disable the border. Otherwise nothing is changed.
+     *
+     * Depending on the theme it might be necessary to set different
+     * CSS rules, so it is not possible to find a solution that fits
+     * everyone. The following configuration works well with the GNOME
+     * default theme Adwaita, but for example has problems under Ubuntu.
+     * Note that for bigger modifications the user can create a style.css
+     * file in tildas config directory, which will be parsed by
+     * load_custom_css_file() in tilda.c on start up.
+     */
     gtk_notebook_set_show_tabs (GTK_NOTEBOOK(tw->notebook), FALSE);
-    gtk_notebook_set_show_border (GTK_NOTEBOOK (tw->notebook), config_getbool("notebook_border"));
+    gtk_notebook_set_show_border (GTK_NOTEBOOK (tw->notebook),
+        config_getbool("notebook_border"));
     tilda_window_set_tab_position (tw, config_getint ("tab_pos"));
 
+    provider = gtk_css_provider_new ();
+    style_context = gtk_widget_get_style_context(tw->notebook);
+    gtk_style_context_add_provider (style_context,
+        GTK_STYLE_PROVIDER(provider),
+        GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
+
+    if(!config_getbool("notebook_border")) {
+        /**
+         * Calling gtk_notebook_set_show_border is not enough. We need to
+         * disable the border explicitly by using CSS.
+         */
+        gtk_css_provider_load_from_data (GTK_CSS_PROVIDER (provider),
+                                " .notebook {\n"
+                                "   border: none;\n"
+                                "}\n", -1, NULL);
+    }
+
+    g_object_unref (provider);
+
     /* Create the linked list of terminals */
     tw->terms = NULL;
 
@@ -410,10 +618,11 @@ tilda_window *tilda_window_init (const gchar *config_file, const gint instance)
     }
 
     /* Connect signal handlers */
-    g_signal_connect (G_OBJECT(tw->window), "delete_event", GTK_SIGNAL_FUNC(gtk_main_quit), tw->window);
-    g_signal_connect (G_OBJECT(tw->window), "show", GTK_SIGNAL_FUNC(focus_term), tw->notebook);
-
-    g_signal_connect (G_OBJECT(tw->window), "focus-out-event", GTK_SIGNAL_FUNC(focus_out_event_cb), tw->window);
+    g_signal_connect (G_OBJECT(tw->window), "delete-event", G_CALLBACK (delete_event_callback), tw);
+    g_signal_connect (G_OBJECT(tw->window), "show", G_CALLBACK (focus_term), tw);
+    g_signal_connect (G_OBJECT(tw->window), "focus-out-event", G_CALLBACK (focus_out_event_cb), tw);
+    g_signal_connect (G_OBJECT(tw->window), "enter-notify-event", G_CALLBACK (mouse_enter), tw);
+    g_signal_connect (G_OBJECT(tw->window), "leave-notify-event", G_CALLBACK (mouse_leave), tw);
 
     /* Add the notebook to the window */
     gtk_container_add (GTK_CONTAINER(tw->window), tw->notebook);
@@ -424,9 +633,11 @@ tilda_window *tilda_window_init (const gchar *config_file, const gint instance)
 
     /* Position the window */
     tw->current_state = UP;
-    gtk_window_move (GTK_WINDOW(tw->window), config_getint ("x_pos"), config_getint ("y_pos"));
     gtk_window_set_default_size (GTK_WINDOW(tw->window), config_getint ("max_width"), config_getint ("max_height"));
     gtk_window_resize (GTK_WINDOW(tw->window), config_getint ("max_width"), config_getint ("max_height"));
+
+	/* Create GDK resources now, to prevent crashes later on */
+    gtk_widget_realize (tw->window);
     generate_animation_positions (tw);
 
     return tw;
@@ -442,14 +653,12 @@ gint tilda_window_free (tilda_window *tw)
     {
         /* Close the 0th tab, which should always exist while we have
          * some pages left in the notebook. */
-        tilda_window_close_tab (tw, 0);
+        tilda_window_close_tab (tw, 0, TRUE);
 
         num_pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK(tw->notebook));
     }
 
     g_free (tw->config_file);
-    g_free (tw->home_dir);
-
     g_free (tw);
 
     return 0;
@@ -484,9 +693,9 @@ gint tilda_window_add_tab (tilda_window *tw)
     /* Create page and append to notebook */
     label = gtk_label_new ("Tilda");
     /* Strangely enough, prepend puts pages on the end */
-    index = gtk_notebook_prepend_page (GTK_NOTEBOOK(tw->notebook), tt->hbox, label);
-    gtk_notebook_set_tab_label_packing (GTK_NOTEBOOK(tw->notebook), tt->hbox, TRUE, TRUE, GTK_PACK_END);
+    index = gtk_notebook_append_page (GTK_NOTEBOOK(tw->notebook), tt->hbox, label);
     gtk_notebook_set_current_page (GTK_NOTEBOOK(tw->notebook), index);
+	gtk_notebook_set_tab_reorderable(GTK_NOTEBOOK(tw->notebook), tt->hbox, TRUE);
 
     /* We should show the tabs if there are more than one tab in the notebook */
     if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (tw->notebook)) > 1)
@@ -509,7 +718,7 @@ gint tilda_window_add_tab (tilda_window *tw)
  * Success: return 0
  * Failure: return non-zero
  */
-gint tilda_window_close_tab (tilda_window *tw, gint tab_index)
+gint tilda_window_close_tab (tilda_window *tw, gint tab_index, gboolean force_exit)
 {
     DEBUG_FUNCTION ("tilda_window_close_tab");
     DEBUG_ASSERT (tw != NULL);
@@ -534,9 +743,50 @@ gint tilda_window_close_tab (tilda_window *tw, gint tab_index)
     if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (tw->notebook)) == 1)
         gtk_notebook_set_show_tabs (GTK_NOTEBOOK (tw->notebook), FALSE);
 
-    /* With no pages left, it's time to leave the program */
-    if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (tw->notebook)) < 1)
-        gtk_main_quit ();
+    /* With no pages left, either leave the program or create a new
+     * terminal */
+    if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (tw->notebook)) < 1) {
+        if (force_exit == TRUE) {
+            /**
+             * It is necessary to check the main_level because if CTRL_C was used
+             * or the "Quit" option from the context menu then gtk_main_quit has
+             * already been called and cannot call it again.
+             */
+            if(gtk_main_level () > 0) {
+                gtk_main_quit ();
+            }
+        } else {
+            /* These can stay here. They don't need to go into a header
+             * because they are only used at this point in the code. */
+            enum on_last_terminal_exit { EXIT_TILDA,
+                                         RESTART_TERMINAL,
+                                         RESTART_TERMINAL_AND_HIDE };
+
+            /* Check the user's preference for what to do when the last
+             * terminal is closed. Take the appropriate action */
+            switch (config_getint ("on_last_terminal_exit"))
+            {
+                case RESTART_TERMINAL:
+                    tilda_window_add_tab (tw);
+                    break;
+                case RESTART_TERMINAL_AND_HIDE:
+                    tilda_window_add_tab (tw);
+                    pull (tw, PULL_UP);
+                    break;
+                case EXIT_TILDA:
+                default:
+                    /**
+                     * It is necessary to check the main_level because if CTRL_C was used
+                     * or the "Quit" option from the context menu then gtk_main_quit has
+                     * already been called and cannot call it again.
+                     */
+                    if(gtk_main_level () > 0) {
+                        gtk_main_quit ();
+                    }
+                    break;
+            }
+        }
+    }
 
     /* Remove the tilda_term from the list of terminals */
     tw->terms = g_list_remove (tw->terms, tt);
@@ -546,4 +796,3 @@ gint tilda_window_close_tab (tilda_window *tw, gint tab_index)
 
     return 0;
 }
-
